import traceback
import click
from click.decorators import pass_context
from flask import current_app
from flask.cli import with_appcontext
from importlib.metadata import EntryPoint, entry_points

from flaschengeist import logger
from flaschengeist.controller import pluginController
from werkzeug.exceptions import NotFound


@click.group()
def plugin():
    pass


@plugin.command()
@click.argument("plugin", nargs=-1, required=True, type=str)
@with_appcontext
@pass_context
def enable(ctx, plugin):
    """Enable one or more plugins"""
    for name in plugin:
        click.echo(f"Enabling {name}{'.'*(20-len(name))}", nl=False)
        try:
            pluginController.enable_plugin(name)
            click.secho(" ok", fg="green")
        except NotFound:
            click.secho(" not installed / not found", fg="red")


@plugin.command()
@click.argument("plugin", nargs=-1, required=True, type=str)
@with_appcontext
@pass_context
def disable(ctx, plugin):
    """Disable one or more plugins"""
    for name in plugin:
        click.echo(f"Disabling {name}{'.'*(20-len(name))}", nl=False)
        try:
            pluginController.disable_plugin(name)
            click.secho(" ok", fg="green")
        except NotFound:
            click.secho(" not installed / not found", fg="red")


@plugin.command()
@click.argument("plugin", nargs=-1, type=str)
@click.option("--all", help="Install all enabled plugins", is_flag=True)
@with_appcontext
@pass_context
def install(ctx: click.Context, plugin, all):
    """Install one or more plugins"""
    all_plugins = entry_points(group="flaschengeist.plugins")
    if all:
        plugins = [ep.name for ep in all_plugins]
    elif len(plugin) > 0:
        plugins = plugin
        for name in plugin:
            if not all_plugins.select(name=name):
                ctx.fail(f"Invalid plugin name, could not find >{name}<")
    else:
        ctx.fail("At least one plugin must be specified, or use `--all` flag.")

    for name in plugins:
        click.echo(f"Installing {name}{'.'*(20-len(name))}", nl=False)
        try:
            pluginController.install_plugin(name)
        except Exception as e:
            click.secho(" failed", fg="red")
            if logger.getEffectiveLevel() > 10:
                ctx.fail(f"[{e.__class__.__name__}] {e}")
            else:
                ctx.fail(traceback.format_exc())
        else:
            click.secho(" ok", fg="green")


@plugin.command()
@click.argument("plugin", nargs=-1, required=True, type=str)
@with_appcontext
@pass_context
def uninstall(ctx: click.Context, plugin):
    """Uninstall one or more plugins"""

    plugins = {plg.name: plg for plg in pluginController.get_installed_plugins() if plg.name in plugin}
    try:
        for name in plugin:
            pluginController.disable_plugin(plugins[name])
            if (
                click.prompt(
                    "You are going to uninstall:\n\n"
                    f"\t{', '.join([plugin_name for plugin_name in plugins.keys()])}\n\n"
                    "Are you sure?",
                    default="n",
                    show_choices=True,
                    type=click.Choice(["y", "N"], False),
                ).lower()
                != "y"
            ):
                ctx.exit()
            click.echo(f"Uninstalling {name}{'.'*(20-len(name))}", nl=False)
            pluginController.uninstall_plugin(plugins[name])
            click.secho(" ok", fg="green")
    except KeyError:
        ctx.fail(f"Invalid plugin ID, could not find >{name}<")


@plugin.command()
@click.option("--enabled", "-e", help="List only enabled plugins", is_flag=True)
@click.option("--no-header", "-n", help="Do not show header", is_flag=True)
@with_appcontext
def ls(enabled, no_header):
    def plugin_version(p):
        if isinstance(p, EntryPoint):
            return p.dist.version
        return p.version

    plugins = entry_points(group="flaschengeist.plugins")
    installed_plugins = {plg.name: plg for plg in pluginController.get_installed_plugins()}
    loaded_plugins = current_app.config["FG_PLUGINS"].keys()

    if not no_header:
        print(f"{' '*13}{'name': <20}|   version   | {' ' * 8} state")
        print("-" * 63)
    for plugin in plugins:
        is_installed = plugin.name in installed_plugins.keys()
        is_enabled = is_installed and installed_plugins[plugin.name].enabled
        if enabled and is_enabled:
            continue
        print(f"{plugin.name: <33}|{plugin_version(plugin): >12} |  ", end="")
        if is_enabled:
            if plugin.name in loaded_plugins:
                print(click.style("      enabled", fg="green"))
            else:
                print(click.style("(failed to load)", fg="red"))
        elif is_installed:
            print(click.style("     disabled", fg="yellow"))
        else:
            print("not installed")
    for name, plugin in installed_plugins.items():
        if plugin.enabled and name not in loaded_plugins:
            print(f"{name: <33}|{'': >12} |" f"{click.style(' failed to load', fg='red')}")
